import 'dart:typed_data';
import './reader_seeker.dart';

class ImageNtHeaders32 {
  int signature = 0;
  late ImageFileHeader fileHeader;
  late ImageOptionalHeader32 optionalHeader; 

  ImageNtHeaders32(ReaderSeeker readerSeeker){
    var buffer = readerSeeker.getData(24);
    signature = buffer.getUint32(0,Endian.little);
    fileHeader = ImageFileHeader(buffer);
    optionalHeader = ImageOptionalHeader32(readerSeeker);
  }

  @override
  String toString(){
    return '{"signature":${signature.toRadixString(16)}, "fileHeader":${fileHeader.toString()}, "optionalHeader":${optionalHeader.toString()}}';
  }
}

class ImageFileHeader {
  int machine = 0;
  int numberOfSections = 0; 
  int timeDateStamp = 0;
  int pointerToSymbolTable = 0;
  int numberOfSymbols = 0;
  int sizeOfOptionalHeader = 0;
  int characteristics = 0;
  ImageFileHeader(ByteData buffer){
    machine = buffer.getUint16(4,Endian.little);
    numberOfSections = buffer.getUint16(6,Endian.little);
    timeDateStamp = buffer.getUint32(8,Endian.little);
    pointerToSymbolTable = buffer.getUint32(12,Endian.little);
    numberOfSymbols = buffer.getUint32(16,Endian.little);
    sizeOfOptionalHeader = buffer.getUint16(20,Endian.little);
    characteristics = buffer.getUint16(22,Endian.little);
  }

  @override
  String toString(){
    return '{"machine":${machine.toRadixString(16)}, "numberOfSections":${numberOfSections.toRadixString(16)}, "timeDateStamp":${timeDateStamp.toRadixString(16)}, "pointerToSymbolTable":${pointerToSymbolTable.toRadixString(16)}, "numberOfSymbols":${numberOfSymbols.toRadixString(16)}, "sizeOfOptionalHeader":${sizeOfOptionalHeader.toRadixString(16)}, "characteristics":${characteristics.toRadixString(16)}}';
  }
}

class ImageOptionalHeader32 {
  int magic = 0;
  int majorLinkerVersion = 0;
  int minorLinkerVersion = 0;
  int sizeOfCode = 0;
  int sizeOfInitializedData = 0;
  int sizeOfUninitializedData = 0;
  int addressOfEntryPoint = 0;
  int baseOfCode = 0;
  int baseOfData = 0;
  int imageBase = 0;
  int sectionAlignment = 0;
  int fileAlignment = 0;
  int majorOperatingSystemVersion = 0;
  int minorOperatingSystemVersion = 0;
  int majorImageVersion = 0;
  int minorImageVersion = 0;
  int majorSubsystemVersion = 0;
  int minorSubsystemVersion = 0;
  int win32VersionValue = 0; 
  int sizeOfImage = 0;
  int sizeOfHeaders = 0;
  int checkSum = 0;
  int subsystem = 0;
  int dllCharacteristics = 0;
  int sizeOfStackReserve = 0;
  int sizeOfStackCommit = 0;
  int sizeOfHeapReserve = 0;
  int sizeOfHeapCommit = 0;
  int loaderFlags = 0;
  int numberOfRvaAndSizes = 0;
  late List<ImageDataDirectory> dataDirectory; 

  bool is32 = false;

  void _parse32(ByteData buffer){
    is32 = true;
    majorLinkerVersion = buffer.getUint8(0);
    minorLinkerVersion = buffer.getUint8(1);
    sizeOfCode = buffer.getUint32(2,Endian.little);
    sizeOfInitializedData = buffer.getUint32(6,Endian.little);
    sizeOfUninitializedData = buffer.getUint32(10,Endian.little);
    addressOfEntryPoint = buffer.getUint32(14,Endian.little);
    baseOfCode = buffer.getUint32(18,Endian.little);
    baseOfData = buffer.getUint32(22,Endian.little);

    imageBase = buffer.getUint32(26,Endian.little);
    sectionAlignment = buffer.getUint32(30,Endian.little);
    fileAlignment = buffer.getUint32(34,Endian.little);
    majorOperatingSystemVersion = buffer.getUint16(38,Endian.little);
    minorOperatingSystemVersion = buffer.getUint16(40,Endian.little);
    majorImageVersion = buffer.getUint16(42,Endian.little);
    minorImageVersion = buffer.getUint16(44,Endian.little);
    majorSubsystemVersion = buffer.getUint16(46,Endian.little);
    minorSubsystemVersion = buffer.getUint16(48,Endian.little);
    win32VersionValue = buffer.getUint32(50,Endian.little);
    sizeOfImage = buffer.getUint32(54,Endian.little);
    sizeOfHeaders = buffer.getUint32(58,Endian.little);
    checkSum = buffer.getUint32(62,Endian.little);
    subsystem = buffer.getUint16(66,Endian.little);
    dllCharacteristics = buffer.getUint16(68,Endian.little);
    sizeOfStackReserve = buffer.getUint32(70,Endian.little);
    sizeOfStackCommit = buffer.getUint32(74,Endian.little);
    sizeOfHeapReserve = buffer.getUint32(78,Endian.little);
    sizeOfHeapCommit = buffer.getUint32(82,Endian.little);
    loaderFlags = buffer.getUint32(86,Endian.little);
    numberOfRvaAndSizes = buffer.getUint32(90,Endian.little);
    int n = 0;
    dataDirectory = List<ImageDataDirectory>.empty(growable: true);
    for (int i = 0 ;i < 16;i++){
      var data = ImageDataDirectory();
      data.virtualAddress = buffer.getUint32(94 + n,Endian.little);
      data.size = buffer.getUint32(98 + n,Endian.little);
      dataDirectory.add(data);
      n += 8;
    }
  }

  void _parse64(ByteData buffer){
    is32 = false;
    majorLinkerVersion = buffer.getUint8(0);
    minorLinkerVersion = buffer.getUint8(1);
    sizeOfCode = buffer.getUint32(2,Endian.little);
    sizeOfInitializedData = buffer.getUint32(6,Endian.little);
    sizeOfUninitializedData = buffer.getUint32(10,Endian.little);
    addressOfEntryPoint = buffer.getUint32(14,Endian.little);
    baseOfCode = buffer.getUint32(18,Endian.little);
    imageBase = buffer.getUint64(22,Endian.little);
    sectionAlignment = buffer.getUint32(30,Endian.little);
    fileAlignment = buffer.getUint32(34,Endian.little);
    majorOperatingSystemVersion = buffer.getUint16(36,Endian.little);
    minorOperatingSystemVersion = buffer.getUint16(38,Endian.little);
    majorImageVersion = buffer.getUint16(40,Endian.little);
    minorImageVersion = buffer.getUint16(42,Endian.little);
    majorSubsystemVersion = buffer.getUint16(44,Endian.little);
    minorSubsystemVersion = buffer.getUint16(46,Endian.little);
    win32VersionValue = buffer.getUint32(48,Endian.little);
    sizeOfImage = buffer.getUint32(52,Endian.little);
    sizeOfHeaders = buffer.getUint32(56,Endian.little);
    checkSum = buffer.getUint32(60,Endian.little);
    subsystem = buffer.getUint16(64,Endian.little);
    dllCharacteristics = buffer.getUint16(66,Endian.little);
    sizeOfStackReserve = buffer.getUint64(68,Endian.little);
    sizeOfStackCommit = buffer.getUint64(76,Endian.little);
    sizeOfHeapReserve = buffer.getUint64(84,Endian.little);
    sizeOfHeapCommit = buffer.getUint64(90,Endian.little);
    loaderFlags = buffer.getUint32(98,Endian.little);
    numberOfRvaAndSizes = buffer.getUint32(102,Endian.little);

    int n = 0;
    dataDirectory = List<ImageDataDirectory>.empty(growable: true);
    for (int i = 0 ;i < 16;i++){
      var data = ImageDataDirectory();
      data.virtualAddress = buffer.getUint32(106 + n,Endian.little);
      data.size = buffer.getUint32(110 + n,Endian.little);
      dataDirectory.add(data);
      n += 8;
    }
  }

  ImageOptionalHeader32(ReaderSeeker readerSeeker){
    var buffer = readerSeeker.getData(2);
    magic = buffer.getUint16(0,Endian.little);
    if (magic == 267) {
      _parse32(readerSeeker.getData(222));
    } else if (magic == 523) {
      _parse64(readerSeeker.getData(234));
    }
  }

  @override
  String toString(){
    return  '{"magic":"${magic},"loaderFlags":"${loaderFlags.toString()}"}';
  }
}

class ImageDataDirectory {
  int virtualAddress = 0;
  int size = 0; 
}